#!/usr/bin/env python3
# coding: utf-8
"""
Utils for text parsing, manipulation and normalization.

:author: 2012, Pascal Hartig <phartig@weluse.de>
:license: Apache 2
"""

import re
import string


EMOTICONS = frozenset([':)', ':(', '):', '(:', '(-:', ')-:', ':-)', ':-('])


def normalize_text(text):
    """Formats text to strip unneccesary words, punctuation and whitespace.
    Returns a tokenized list.

    :param text: Text to process.

    >>> text = "ommmmmmg how'r U!? VISI T  <html> <a href='http://google.com'> my</a> site @ http://www.coolstuff.com haha"
    >>> normalize_text(text)
    ['ommg', 'howr', 'visi', 'my', 'site', 'haha']

    >>> normalize_text("FOE JAPAN が粘り強く主張していた避難の権利")
    ['foe', 'japan', '\u304c\u7c98\u308a\u5f37\u304f\u4e3b\u5f35\u3057\u3066\u3044\u305f\u907f\u96e3\u306e\u6a29\u5229']

    >>> normalize_text('no ')
    ['no']

    >>> normalize_text('')
    []
    """

    if not text:
        return []

    # Process in lower case
    text = text.lower()

    patterns = (
        ("@[A-Za-z0-9_]+", ''),     # mentions
        ("#[A-Za-z0-9_]+", ''),     # hash tags
        (r"(\w)\1{2,}", r"\1\1"),   # occurences of more than two consecutive
                                    # repeating characters
        ("<[^<]+?>", ''),           # sgml tags
        ("(http|www)[^ ]*", ''),    # URLs
    )

    for pattern in patterns:
        text = re.sub(pattern[0], pattern[1], text)

    # Temporarily store emoticons before stripping.
    emoticons = set([e for e in EMOTICONS if e in text])

    # Remove puctuation, leading/trailing whitespace
    text = text.translate({ord(x): None for x in string.punctuation}).strip()

    # Reappand emoticons
    text += ' '.join([e for e in emoticons])

    return [w for w in re.split(r'\s', text) if len(w) > 1]


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Utils for extracting and preparing information.

:author: 2012, Pascal Hartig <phartig@weluse.de>
:license: Apache 2
"""

def extract_features(document):
    """Extract features from a document and returns them in a Bag Of Words
    model dict.
    """

    return {word: True for word in document}



"""
Top level classifier based on top of NaiveBayesClassifier that
supports persistance and loading training data from the file system.

:author: 2012, Pascal Hartig <phartig@weluse.de>
:license: Apache 2
"""

import json
from twentiment.naivebayes import NaiveBayesClassifier
from twentiment.extract import extract_features
from twentiment.text import normalize_text


def _extract_documents(tweets, label):
    return [(normalize_text(tweet), label) for tweet in tweets]


class Classifier:

    def __init__(self, naive_bayes_classifier):
        """
        Instantiates a new classifier with a trained naive bayes classifier.
        To train, utilize the :meth:`from_file` or :meth:`from_json` factory
        methods.

        :param naive_bayes_classifier: Trained instance of
            :cls:`~twentiment.naivebayes.NaiveBayesClassifier`.
        """

        self.classifier = naive_bayes_classifier

        def _proxy_classifier_method(method):
            return lambda *a, **kw: getattr(self.classifier, method)(*a, **kw)

        # Proxy some methods to the classifier
        for method in ['prob_classify', 'classify']:
            setattr(self, method, _proxy_classifier_method(method))
            setattr(getattr(self, method), '__doc__',
                    getattr(self.classifier, method).__doc__)

    @classmethod
    def from_file(cls, file):
        """Creates a new instance from the given file handle."""

        return cls.from_json(json.load(file))

    @classmethod
    def from_json(cls, json):
        """Creates a new instance from the given JSON data as dict data
        structure.
        """

        data = json['trainingData']
        pos_tweets = data['positive']
        neg_tweets = data['negative']

        tweets = (_extract_documents(data['positive'], 'positive') +
                  _extract_documents(data['negative'], 'negative'))

        training_set = [(extract_features(doc), label) for (doc, label)
                        in tweets]

        return cls.from_training_set(training_set)

    @classmethod
    def from_training_set(cls, training_set):
        """Creates a new instance from the given training set."""

        classifier = NaiveBayesClassifier.train(training_set)
        return cls(classifier)



#!/usr/bin/env python3

"""
A classifier based on the Naive Bayes algorithm.  In order to find the
probability for a label, this algorithm first uses the Bayes rule to
express P(label|features) in terms of P(label) and P(features|label):

|                       P(label) * P(features|label)
|  P(label|features) = ------------------------------
|                              P(features)

The algorithm then makes the 'naive' assumption that all features are
independent, given the label:

|                       P(label) * P(f1|label) * ... * P(fn|label)
|  P(label|features) = --------------------------------------------
|                                         P(features)

Rather than computing P(featues) explicitly, the algorithm just
calculates the denominator for each label, and normalizes them so they
sum to one:

|                       P(label) * P(f1|label) * ... * P(fn|label)
|  P(label|features) = --------------------------------------------
|                        SUM[l]( P(l) * P(f1|l) * ... * P(fn|l) )


:author: Edward Loper <edloper@gradient.cis.upenn.edu>
:author: Pascal Hartig <phartig@weluse.de>
:url: <http://www.nltk.org/>
:license: Apache 2.0
"""

import logging
from collections import defaultdict
from twentiment.thirdparty.probability import (FreqDist, DictionaryProbDist,
                                               ELEProbDist, sum_logs)


class NaiveBayesClassifier(object):
    """
    A Naive Bayes classifier.  Naive Bayes classifiers are
    paramaterized by two probability distributions:

      - P(label) gives the probability that an input will receive each
        label, given no information about the input's features.

      - P(fname=fval|label) gives the probability that a given feature
        (fname) will receive a given value (fval), given that the
        label (label).

    If the classifier encounters an input with a feature that has
    never been seen with any label, then rather than assigning a
    probability of 0 to all labels, it will ignore that feature.

    The feature value 'None' is reserved for unseen feature values;
    you generally should not use 'None' as a feature value for one of
    your own features.
    """

    LOG = logging.getLogger('NaiveBayesClassifier')

    def __init__(self, label_probdist, feature_probdist):
        self._label_probdist = label_probdist
        self._feature_probdist = feature_probdist
        self._labels = list(label_probdist.samples())

    @staticmethod
    def train(labeled_featuresets, estimator=ELEProbDist):
        """
        :param labeled_featureset: A set of classified featuresets,
            i.e., a list of tuples ``[(featureset, label)]``.
        :param estimator: An estimator probability distribution. Defaults to an
            expected likelyhood estimation probability distribution.
        """

        label_freqdist = FreqDist()
        #: Features and values are stored in dictionaries defaulting to
        #: empty frequency distributions or sets, respectively.
        feature_freqdist = defaultdict(FreqDist)
        feature_values = defaultdict(set)
        #: Set of all feature names used, across labels.
        fnames = set()

        # Count how many times a particular feature value occured given the
        # label and feature name.
        for featureset, label in labeled_featuresets:
            # Track every label occurence.
            label_freqdist.inc(label)

            for fname, fval in featureset.items():
                # Increment the freq(fval|label, fname)
                feature_freqdist[label, fname].inc(fval)
                # Record that the fname can take the value fval.
                feature_values[fname].add(fval)
                # Keep a set of all used feature names.
                fnames.add(fname)

        for label in label_freqdist:
            num_samples = label_freqdist[label]
            for fname in fnames:
                # The count of the feature given the label.
                count = feature_freqdist[label, fname].N()
                # Create a balance between the labels, such that for each
                # 'missing' occasion of this feature in the current label,
                # there is a 'None' element incremented. So in the end all
                # freqdists have the same count, but with different fvalues.
                feature_freqdist[label, fname].inc(None, num_samples - count)
                # Make sure the values are aware of the None value.
                if (num_samples - count > 0):
                    feature_values[fname].add(None)

        #: The distribution P(label)
        label_probdist = estimator(label_freqdist)

        #: The distribution P(fval|label, fname)
        feature_probdist = {}
        for ((label, fname), freqdist) in feature_freqdist.items():
            # Create the estimator with as many bins as there are values of the
            # current feature name.
            probdist = estimator(freqdist, bins=len(feature_values[fname]))
            feature_probdist[label, fname] = probdist

        return NaiveBayesClassifier(label_probdist, feature_probdist)

    def prob_classify(self, featureset):
        """Calculate the probabilities the given featureset classifications
        and return a DictionaryProbDist instance.

        Works in O(nm) with n = # of labels, m = # of featureset elements.
        """

        # Work on a copy of the feature set, because we mutate it.
        fset = featureset.copy()
        for fname in featureset:
            for label in self._labels:
                if (label, fname) in self._feature_probdist:
                    break
            else:
                # Discard feature name we haven't been trained on from the
                # input set.
                del fset[fname]

        # Now we're working with a feature set that only includes known
        # features.

        # Instead of working with the product of the separate probabilities,
        # we use the sum of the logarithms to prevent underflows and make the
        # result more stable.

        #: The probability of each label, given the features. Starting with
        #: the probability of the label itself.
        logprob = {}
        for label in self._labels:
            logprob[label] = self._label_probdist.logprob(label)

        # Add the logarithmic probability of the features given the labels.
        for label in self._labels:
            for (fname, fval) in fset.items():
                feature_probs = self._feature_probdist.get((label, fname))

                if feature_probs is not None:
                    logprob[label] += feature_probs.logprob(fval)
                else:
                    # This should not occur if the classifier was created with
                    # the train() method.
                    logprob[label] += sum_logs([])  # = -INF.

        return DictionaryProbDist(logprob, normalize=True, log=True)

    def classify(self, featureset):
        """Return the most likely label for a given featureset."""

        return self.prob_classify(featureset).max()



"""
A ZeroMQ-based server that answeres guessing queries.

:author: 2012, Pascal Hartig
:license: Apache 2
"""

import zmq
from twentiment.extract import extract_features
from twentiment.text import normalize_text


class Server:
    """A simple ZMQ-based server listening on a customizable bind.

    Protocol*::

        -> GUESS [tweet:str]
        <- OK [guess:float]
        - OR -
        <- ERROR [code:str] [description?:str]

    Possible Errors:

        * UNKNOWN_COMMAND: The leading command was not understood.
        * RUNTIME_ERROR: An error on the server side occured.
        * BAD_FORMAT: A request must start with a command separated by an
            ASCII space (20)

    (* Not really worth calling it that.)
    """
    def __init__(self, classifier, bind="tcp://127.0.0.1:10001"):
        """Creates a new server instance.

        :param bind: The zmq bind, defaults to tcp://127.0.0.1:10001.
            Obviously, the same must be used on the client side.
        """

        self.bind = bind
        self.classifier = classifier

    def run(self):
        """Starts a blocking server."""

        context = zmq.Context()
        socket = context.socket(zmq.REP)

        socket.bind(self.bind)

        print("Starting server on {}".format(self.bind))
        while True:
            message = socket.recv()
            try:
                response = self._handle_message(str(message, "utf-8"))
            except Exception as err:
                response = self._error_response("RUNTIME_ERROR " + str(err))
                socket.send_unicode(response)

                raise

            socket.send_unicode(response)

    def _handle_message(self, message):
        try:
            cmd, message = message.lower().split(" ", 1)
        except ValueError:
            return self._error_response("BAD_FORMAT")

        if cmd == 'guess':
            return "OK {}".format(self._guess(message))
        else:
            return self._error_response("UNKNOWN_COMMAND")

    def _error_response(self, message):
        return "ERROR {}".format(message)

    def _guess(self, message):
        twfeat = extract_features(normalize_text(message))
        result = self.classifier.prob_classify(twfeat)
        return format(result.prob('positive') - result.prob('negative'))



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


import twentiment


setup(
    name='twentiment',
    version=twentiment.__version__,
    description='Twitter sentiment analysis tool',
    long_description=open('README.rst').read(),
    author='Pascal Hartig',
    author_email='phartig@rdrei.net',
    url='https://github.com/passy/twentiment',
    packages=['twentiment', 'twentiment.thirdparty'],
    package_data={'': ['LICENSE', 'README.rst']},
    include_package_data=True,
    scripts=["bin/twentiment_server", "bin/twentiment_client"],
    install_requires=[
        'pyzmq',
        'six==1.2.0'
    ],
    license=open('LICENSE').read(),
    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Operating System :: Unix',
        'Topic :: Communications',
        'Topic :: Internet :: WWW/HTTP'
    )
)
